// CForth  project ************************************************************
//
// Author and (c)2020-24 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.h","cforth.c","cforth.h",
// "usb-otg-ctrlbd.c" and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH/Rowley Associates Ltd
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32l4xx.h"
#include "globaldefines.h"
#include "main.h"

int main(void)    // ********************************************************
// **************************************************************************
{
uint32_t k;

    RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN;  //necessary for what?? 
    RCC->AHB2ENR |= RCC_AHB2ENR_GPIOAEN;  // Enable Port A clock
    RCC->AHB2ENR |= RCC_AHB2ENR_GPIOBEN;  // Enable Port B clock
    RCC->AHB2ENR |= RCC_AHB2ENR_GPIOCEN;  // Enable Port B clock
    RCC->AHB2ENR |= RCC_AHB2ENR_GPIODEN;  // Enable Port B clock

    GPIOA->MODER = 0x00000000;      //primarily all input
    GPIOB->MODER = 0x00000000;      //primarily all input
    GPIOC->MODER = 0x00000000;      //primarily all input
    GPIOD->MODER = 0x00000000;      //primarily all input (physically only PD2)

    GPIOA->PUPDR = 0xA402AAAA;      //primarily all inputs pulldown,
                                    //except PA9,10,11,12,13,14
    GPIOB->PUPDR = 0xAAAAAAAA;      //primarily all inputs pulldown
	GPIOB->PUPDR &= ~(GPIO_PUPDR_PUPD2); //PB2 is Vusb bridge,must be HiZ
    GPIOC->PUPDR = 0xAAAAAAAA;		//primarily all inputs pulldown,
//the DIP switch on the veroboard is not used by CForth
//but it can be used as I/O (PC2, PC3)

//#endif
    GPIOD->PUPDR = 0x00000020;	//primarily PD2 pullDown (only existing one)
    SetSysClock();

#ifdef STLINK_CUTOFF
//pin config repeated specifically:
	GPIOC->PUPDR &= ~(GPIO_PUPDR_PUPD0);	//pull-down off, wasts current
    GPIOC->MODER &= ~(GPIO_MODER_MODE0);
    GPIOC->MODER |= GPIO_MODER_MODE0_0 ;  	//set PortC0 (green LED ) as output
    GPIOC->BSRR =  GPIO_BSRR_BS_0;        	//PC0 (green LED) ON
#else
//pin config repeated specifically:
	GPIOA->PUPDR &= ~(GPIO_PUPDR_PUPD5);	//pull-down off, wasts current
    GPIOA->MODER &= ~(GPIO_MODER_MODE5);
    GPIOA->MODER |= GPIO_MODER_MODE5_0 ;  	//set PortA5 (user LED) as output
    GPIOA->BSRR =  GPIO_BSRR_BS_5;        	//PA5 (user LED) ON
#endif

    gSYSCOUNT = 50;
    gLEDCOUNT = 0;
    gDELAYCOUNT = 0;
    gTIMEOUT = 0;
    gProject = 0;
	gfRandom = 314159265;		//seed for xorshift32 algorithm
	gfNumBase = DEFAULT_NUMBASE;
	gfTempNumBase = DEFAULT_NUMBASE;
    gBaud = DEFAULT_BAUD;
#ifdef STLINK_CUTOFF
    gTempBaud = DEFAULT_BAUD;
    gVid = DEFAULT_VID;
    gPid = DEFAULT_PID;
    guDeviceClass = CDC_CLASS; //seems to be necessary
    guConfigured = 0;
    gUsbChecked = 1;
	gSplit = 0;					//USB and USART in parallel as terminal
	gBaudSet = 0;
	gChannel = 0;
	gMIDI_HEAD = 0;
	gMIDI_TAIL = 0;
#endif

    SysTick_Config(SystemCoreClock/1000);

    msDelay(100);

//start Independent Watchdog  **************
    IWDG->KR = 0x00005555;
    IWDG->RLR = 0x0000FFFF;
    msDelay(2);
    IWDG->PR = 0x00000002;      //resulting timeout ca. 1-2 seconds
    msDelay(2);
    IWDG->KR = 0x0000CCCC;      //start Independent Watchdog

//init I/O state  *************************
    popFlashPage23();		//ifnot STLINK_CUTOFF:contains some dummies
 //init communication  ********************

#ifdef STLINK_CUTOFF
    initUsart1();
#else               //original, un-modified Nucleo board assumed
    initUsart2();
#endif
    msDelay(100);

//popFlashPage23();						uncomment makes sense for debugging

#ifdef STLINK_CUTOFF
	sendModel();
    msDelay(2);
    guConfigured = 0;

    if ((GPIOA->IDR&GPIO_IDR_ID9) && (gVid != 0) && (gPid != 0)) startUSB();
    else  {
        sendB('n');
        sendB('o');
        sendB(' ');
        sendB('U');
        sendB('S');
        sendB('B');
        sendB(CRR);
    }
#else
//file "usb-otg.c"  is included for compiler reasons, but not used in this case
#endif
    popFlashProject();                  //now load complete project

    while(1) {
	quit();
	query();
	if (gfTibMax == 0) continue; 
	eval();
    }
    msDelay(100);
    while(1);       //force cold reset by watchdog
}

//##########################################################################

// Misc Low Level Routines
//************************

void sendModel(void)  //  ****************************************************
//  **************************************************************************
{

const uint8_t typeStr [] = "CForth for STM32 Nucleo-L476 v";
const uint8_t copyStr [] = "  (c)2020-24 W.Schemmert";

	sendTB(CRR);
	sendString(typeStr);
	sendTB(HARDWARE_VERSION + 0x30);
	sendTB('.');
	sendNS32(REVNUM);
	sendString(copyStr);
	sendTB(CRR);
}


void SetSysClock(void)  // *******************************************
//MSI (48MHz) is stabilized by PLL with LSE
//provides "Vcore range 1", MSI clock 48.000 MHz, no PLL for SYSCLK etc,
//AHB=SystemClock, APB1+2 = 48 MHz
// **************************************************************************
{

//enable power to the backup domain (LSE oscillator)
    RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN;
//adjust Vcore Range1 and disable backup domain write protection
    while (PWR->SR2 & PWR_SR2_VOSF) ;
    PWR->CR1 &= ~PWR_CR1_VOS_1;
    PWR->CR1 |= (PWR_CR1_VOS_0 | PWR_CR1_DBP);
    while (PWR->SR2 & PWR_SR2_VOSF) ;

//adjust Flash Waitstates and Prefetch for higher System Clock frequency
    FLASH->ACR |= (FLASH_ACR_PRFTEN | FLASH_ACR_LATENCY_2WS); //for SYSCLK=48MHz

    RCC->CR &= ~RCC_CR_PLLON;           //stop PLL, MSI keeps running with default 4MHz
    while(RCC->CR & RCC_CR_PLLRDY) ;    //as a measure of precaution

    RCC->BDCR &= ~0x0000003D;           //reset all LSE relevant bits
//start LSE ocscillator
 //   RCC->BDCR  &= ~RCC_BDCR_LSEDRV;
    RCC->BDCR |= (RCC_BDCR_LSEDRV_0 | RCC_BDCR_LSEDRV_1);     //high drive
    RCC->BDCR |= RCC_BDCR_LSEON;        //start
    while (!(RCC->BDCR & RCC_BDCR_LSERDY));

//adjust MSI oscillator
    while (!(RCC->CR & RCC_CR_MSIRDY)); //as precaution, should be ok
    RCC->CR &= ~RCC_CR_MSIRANGE;        //clear all relevant bits
    RCC->CR |= RCC_CR_MSIRANGE_11;      //set new value (ca.48MHz)
    RCC->CR |= (RCC_CR_MSIRGSEL | RCC_CR_MSIPLLEN | RCC_CR_MSION);
										//synchronize MSI with stable LSE
    while (!(RCC->CR & RCC_CR_MSIRDY)); //confirm PLL lockin at new range

//SYSCLK=MSI and HCLK = SYSCLK, already set in "SystemInit()"
    RCC->CFGR |= (RCC_CFGR_PPRE1_DIV1 | RCC_CFGR_PPRE2_DIV1);
// PCLK1 = HCLK = 48MHz
// PCLK2 = HCLK = 48MHz

// select MSI as clock for USB
    RCC->CCIPR |= RCC_CCIPR_CLK48SEL;    //set CLK48SEL[1:0]=11

    SystemCoreClock = 48000000;

}


void LED_Off(void)  // ******************************************************
// built as subroutine to supply service to "usb-fs-431.c" hardware independent
// **************************************************************************
{
#ifdef STLINK_CUTOFF
    GPIOC->BSRR = GPIO_BSRR_BR_0; // clear bit PC0,LED OFF
#else
    GPIOA->BSRR = GPIO_BSRR_BR_5; // clear bit PA5,LED OFF
#endif
    gLEDCOUNT = 4; //ca. 300 milliseconds
}


void msDelay(uint32_t count)   // *******************************************
//blocking delay by 'count' milliseconds
// **************************************************************************
{
    gDELAYCOUNT = count;
	while(gDELAYCOUNT)  servSystem();
}


void wait5us(void)   // *****************************************************
// blocking delay ca 5us
// *************************************************************************
{
uint32_t countdown;

	countdown = 51;
    while (countdown)   {
        IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
        countdown-- ;
    }
}


void servSystem(void)   // ************************************************
//not very time critical but permanently repeated functions for system control
//exclusively called in pullUsartRxBuf and pullUsbRxBuf
//this way it is regularily served when idle waiting for external input
//but watchdog resets when it hangs somewhere else
//additionally slowed down by SysTick to 50 times/second
// ****************************************************************************
{
static uint32_t secCount;
const uint8_t plugStr [] = "USB plugged ... " ;
const uint8_t unplugStr [] = "USB unplugged" ;

	if (gSYSCOUNT)  return;		//countdown in SysTick_Handler()

//else if gSYSCOUNT == 0
    gSYSCOUNT = 50;				//restart 20Hz service interval
    IWDG->KR = 0x0000AAAA;		//buy time:feed watchdog,timeout ca.1-2seconds

#ifdef STLINK_CUTOFF
//LED control
	if (gLEDCOUNT) gLEDCOUNT--;
    if (gLEDCOUNT==0)  GPIOC->BSRR =  GPIO_BSRR_BS_0; //PC0(User LED) ON again

//USB mode control
    secCount++;
    if (secCount >= 20)  {  		//restart 1 second service interval
        secCount = 0;

        if (guConfigured == 0) {
            if ((GPIOA->IDR & GPIO_IDR_ID9) && (gUsbChecked == 0)) {
                sendString(plugStr);
                startUSB();       	//USB plugged, enforce reset by Watchdog
            }
        }
        else  {
            if ((GPIOA->IDR & GPIO_IDR_ID9) == 0)  {
                sendString(unplugStr);
                sendB(CRR);
                stopUSB();         	//USB un-plugged, deactivate peripheral
                gUsbChecked = 0;
            }
			else   {                //USB plugged
				if ((gSplit == 4) && ((guDeviceClass) == CDC_CLASS))  {
					stopUSB();
					guDeviceClass = MIDI_CLASS;	//added 23Oct21
					msDelay(500);
					startUSB();
				}
				if ((gSplit < 4) && ((guDeviceClass)== MIDI_CLASS))  {
					stopUSB();
					guDeviceClass = CDC_CLASS;	//added 23Oct21
					msDelay(500);
					startUSB();
				}
			}
        }
    }
#else
//LED control
	if (gLEDCOUNT) gLEDCOUNT--;
    if (gLEDCOUNT==0)  GPIOA->BSRR =  GPIO_BSRR_BS_5; //PA5 (User LED) ON again
#endif
}


#ifdef STLINK_CUTOFF
void startUSB(void)  // ******************************************************
// ***************************************************************************
{
uint32_t oldTimeout;
const uint8_t cdcClassStr [] = "USB as virtual COM =" ;
const uint8_t midiClassStr [] = "USB as MIDI interface =" ;

    gUsbChecked = 1;
    if (gSplit == 4) guDeviceClass = MIDI_CLASS; //double safe init
    else guDeviceClass = CDC_CLASS;
    initUSB();
    gTIMEOUT = 3000;
	oldTimeout = 3000;
	while (gTIMEOUT)  {
		IWDG->KR = 0x0000AAAA;		//feed watchdog, timeout ca.0.5c
		if (gTIMEOUT != oldTimeout)  {	//check once per ms
			oldTimeout = gTIMEOUT;
			if ((gTIMEOUT & 0x00000003FF) == 0x3FF)  sendTB('.');
		}
        if (guConfigured)  break;
    }
    if (guDeviceClass == MIDI_CLASS)  sendString(midiClassStr);
    else  sendString(cdcClassStr);	
    if (guConfigured)  {
        sendTB('o');
        sendTB('k');
    }
    else {
        sendB('e');
        sendB('r');
        sendB('r');
        sendB('o');
        sendB('r');
    }
    sendTB(CRR);
    msDelay(200);
}
#endif  //STLINK_CUTOFF

void popFlashPage23(void)   // **********************************************
//read project independent system parameters
// **************************************************************************
{
uint32_t * flashPtr;
uint32_t test, vid, pid;
uint32_t *sramPtr;

	emptyForth();           //init default system paramters

    flashPtr = (uint32_t *)PAGE23;

    test = ~(*flashPtr);//virgin test cannot be==0 if project was saved in Flash
    flashPtr++;
    vid =  ~(*flashPtr);    //cannot be==0 if any project was saved in Flash
    flashPtr++;
    pid =  ~(*flashPtr);    //cannot be==0 if any project was saved in Flash
    flashPtr++;
    if ((test | vid | vid)==0) {//Flash is virgin, load default empty project
#ifdef STLINK_CUTOFF
		gBaud = DEFAULT_BAUD;	//as precaution, should be already set
		gTempBaud = DEFAULT_BAUD;
#endif
		return;					//defaults already init
    }
#ifdef STLINK_CUTOFF
    gVid = vid;
    gPid = pid;
	test =  ~(*flashPtr);
	flashPtr++;
	gBaud = test & 0x0000007F;	//baud rate is project independent
	if (gBaud == 0) gBaud = DEFAULT_BAUD;
	gTempBaud = gBaud;
	gfNumBase = (test >> 16) & 0x0000001F;
	if ((gfNumBase!=10)&&(gfNumBase!=16)) gfNumBase = DEFAULT_NUMBASE;
    gChannel = ~(*flashPtr);    //MIDI channel is project independent
    flashPtr++;
#else
    flashPtr++;
    flashPtr++;
#endif
    gProject =  ~(*flashPtr);   //number of project which is loaded by default
								//gProject=0 loads Empty Project
//	if (gProject == 0)  emptyForth();	//bugfix 25Jan24
}


uint32_t pushFlashPage23(void)   //******************************************
//save project independent system parameters permanently
// **************************************************************************
{
uint32_t * flashPtr;
uint32_t word;

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
    else return 0;
 
//feed watchdog
    IWDG->KR = 0x0000AAAA;          //timeout ca. 1-2 seconds

//erase Page23
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = ((23<<3) | FLASH_CR_EOPIE | FLASH_CR_ERRIE | FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write mirror into Page23:
//this microcontroller allows exclusively doubleword-wide write in flash
//flash least word is first word in readout
//only write USB Vid/Pid, gBaud and gProject into page23
    flashPtr = (uint32_t *)PAGE23;

    FLASH->CR |= FLASH_CR_PG;
#ifdef STLINK_CUTOFF
    *flashPtr = 0;                 //for flash virgin test
    flashPtr++;
    *flashPtr = ~(gVid & 0x0000FFFF);
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    *flashPtr = ~(gPid & 0x0000FFFF);
    flashPtr++;

	*flashPtr = ~((gBaud & 0x0000007F) | 
				(gfNumBase << 16));
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    *flashPtr = ~gChannel;
    flashPtr++;
    *flashPtr = ~gProject;
    flashPtr++;
#else		//dummy entries for compatibility of Flash with STLINK_CUTOFF
    *flashPtr = 0;                 //for flash virgin test
    flashPtr++;
    *flashPtr = ~(DEFAULT_VID & 0x0000FFFF);
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    *flashPtr = ~(DEFAULT_PID & 0x0000FFFF);
    flashPtr++;
    *flashPtr = ~DEFAULT_BAUD;
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    *flashPtr = 0xFFFFFFFF;		//dummy for gChannel
    flashPtr++;
    *flashPtr = ~gProject;
    flashPtr++;
#endif
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR) return 0;

//terminate
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
    return  1;
}


void popFlashProject(void)   // *********************************************
//first read some project specific globals
//next read 2 pages to gfUserHeader[], then read 6 pages to gfTCode
// **************************************************************************
{
uint32_t k, m, n;
uint32_t * flashPtr;
uint32_t word, test;

//emptyForth() already updated in popFlashPage23() (default system parameters)

	if (gProject)  {
		flashPtr = (uint32_t *)(PAGE24 + 0x4000 * (gProject - 1));
//first 32 bytes with system & peripheral config parameters
		test =  ~(*flashPtr);	//cannot be==0 if any project was saved in Flash
		flashPtr++;
		if (test == 0) {		//project Flash is virgin
			emptyForth();		//implies gProject=0 -> empty project
		}
		else  {
#ifdef STLINK_CUTOFF
			gMIDI_HEAD = 0;			//flush MIDI buffer
			gMIDI_TAIL = 0;
#endif
			gfNumBase= ~(*flashPtr);
			flashPtr++;
			if ((gfNumBase!=10)&&(gfNumBase!=16)) gfNumBase = DEFAULT_NUMBASE;
			gfTempNumBase = gfNumBase;
			gfAuto = ~(*flashPtr);
			flashPtr ++;

			word = ~(*flashPtr);
			flashPtr ++;
			gfSpiMode = (word >> 24) & 0x00000003;
			gfSpiBits = (word >> 16) & 0x00000003F;
			gfSpiClock = word & 0x00000FFFF;

			gfFreq12 = ~(*flashPtr);
			flashPtr ++;
			k = ~(*flashPtr);
			gfPWM1 = (int32_t)(k >> 16);
			if ((gfPWM1 == 0) || (gfPWM1 == 0x0000FFFF)) gfPWM1 = -1;
			gfPWM2 = (int32_t)(k & 0x0000FFFF);
			if ((gfPWM2 == 0) || (gfPWM2 == 0x0000FFFF)) gfPWM2 = -1;
			flashPtr ++;
			if (gfFreq12)  {
				subFREQ12();
				subPWM1();
				subPWM2();
			}
			gfFreq3 = ~(*flashPtr);
			flashPtr ++;
			k = ~(*flashPtr);
			gfPWM3 = (int32_t)(k >> 16);
			if ((gfPWM3 == 0) || (gfPWM3 == 0x0000FFFF)) gfPWM3 = -1;
			flashPtr ++;
			if (gfFreq3)  {
				subFREQ3();
				subPWM3();
			}

			for (k = 0;k < MAXUSER/2;k++)  {
									//reads 84*24 words=2016bytes +32=2048
				n = 0;
				for (m = 0; m < 4; m++) {
					word = ~(*flashPtr);
					flashPtr++;
					gfUserHeader[k].name[n]= word & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 8) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 16) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 24) & 0x000000FF;
					n++;
				}
				gfUserHeader[k].toCompile= ~(*flashPtr);
				flashPtr ++;
				gfUserHeader[k].attributes = ~(*flashPtr);
				flashPtr ++;
			}
			flashPtr = (uint32_t *)(PAGE25 + 0x4000 * (gProject - 1));
			for (k = (MAXUSER/2) +1;k < MAXUSER;k++) {  //reads 2044 bytes!
				n = 0;
				for (m = 0; m < 4; m++) {
					word = ~(*flashPtr);
					flashPtr++;
					gfUserHeader[k].name[n]= word & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 8) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 16) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 24) & 0x000000FF;
					n++;
				}
				gfUserHeader[k].toCompile= ~(*flashPtr);
				flashPtr ++;
				gfUserHeader[k].attributes = ~(*flashPtr);
				flashPtr ++;
			}
	
			flashPtr = (uint32_t *)(PAGE26 + 0x4000 * (gProject - 1));
			for (k = 0;k < MAXCODE;k++)  {	//reads 6 pages each 2048 bytes!
				gfTCode[k] = ~(*flashPtr);
				flashPtr ++;
			}
		}
	}
//from here if(gProject) or virgin  or not:
	wait5us();
    op_ABORT();
	if (gfAuto)  {
//emergency brake to avoid AUTO deadlock
		if ((GPIOC->IDR & 0x2000)==0) gfAuto =0; //if PC13 (user button)is LOW
	}

    op_QUEST();
    if (gfAuto) {
        gfParsed = gfAuto - 1;	//-1 to distinguish from empty AUTOEXE == 0
        interpretUserToken();
    }
}


uint32_t pushFlashProject(void)   //*****************************************
// **************************************************************************
{
uint32_t k, m, n, offset;
uint32_t * flashPtr;
uint32_t word[12];

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    if (gProject == 0) return 1;   //=default empty project, don't save

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

//feed watchdog
    IWDG->KR = 0x0000AAAA;          //timeout ca. 1-2 seconds

    offset = (gProject - 1) * 8;     //project no.0 is empty, not treated here
                                    //project no.1 starts with page#24
//erase Page24 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((24 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//this microcontroller allows exclusively doubleword-wide write in flash
//first write virgin test, relocalisation, gfNumBase and gfAuto into Page32

    flashPtr = (uint32_t *)(PAGE24 + 0x800 * offset); //0x4000*(project-1));
    FLASH->CR |= FLASH_CR_PG;
//first 32 bytes with system & peripheral config parameters
    *flashPtr = 0x00000000; //virgin flash test
    flashPtr++;
    *flashPtr = ~gfNumBase;
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR) return 0;

    *flashPtr = ~gfAuto;
    flashPtr++;
    *flashPtr = (~((gfSpiMode << 24) | (gfSpiBits << 16 ) |gfSpiClock));
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    *flashPtr = ~gfFreq12;
	flashPtr++;
    *flashPtr = ~((((uint32_t)gfPWM1 << 16) & 0xFFFF0000) | 
											((uint32_t)gfPWM2 & 0x0000FFFF));
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    *flashPtr = ~gfFreq3;
	flashPtr++;
    *flashPtr = ~((((uint32_t)gfPWM3 << 16) & 0xFFFF0000));
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

//write first half of gfUserHeader[] into Page24 + offset
    for (k = 0;k < MAXUSER/2; k+=2) {	//MAXUSER/2 = 84 ->2016bytes+32=2048
        n = 0;
        for (m = 0; m < 4; m++) {
            word[m] = gfUserHeader[k].name[n];
            n++;
            word[m] |= gfUserHeader[k].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 24;
            n++;
        }
        word[4] = gfUserHeader[k].toCompile;
        word[5] = gfUserHeader[k].attributes;

        n = 0;
        for (m = 6; m < 10; m++) {
            word[m] = gfUserHeader[k+1].name[n];
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 24;
            n++;
        }
        word[10] = gfUserHeader[k+1].toCompile;
        word[11] = gfUserHeader[k+1].attributes;

        for (m = 0; m < 12; m+=2)  {
// this microcontroller allows exclusively doubleword-wide write in flash
//first word
            *flashPtr = ~word[m]; //~for better compatibility with empty SRAM
            flashPtr++;
//second word
            *flashPtr = ~word[m+1]; //~for better compatibility with empty SRAM
            flashPtr++;
            while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
            FLASH->SR |= FLASH_SR_EOP;
            if (FLASH->SR & FLASH_SR_OPERR)  return 0;
        }
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

//erase Page25 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((25 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write second half of gfUserHeader[] into Page25 + offset
    flashPtr = (uint32_t *)(PAGE25 + 0x800 * offset); //0x4000 * (project - 1));

    FLASH->CR |= FLASH_CR_PG;
    for (k = (MAXUSER/2);k < MAXUSER; k+=2) {     //MAXUSER/2 = 84
        n = 0;
        for (m = 0; m < 4; m++) {
            word[m] = gfUserHeader[k].name[n];
            n++;
            word[m] |= gfUserHeader[k].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 24;
            n++;
        }
        word[4] = gfUserHeader[k].toCompile;
        word[5] = gfUserHeader[k].attributes;
        n = 0;
        for (m = 6; m < 10; m++) {
            word[m] = gfUserHeader[k+1].name[n];
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 24;
            n++;
        }
        word[10] = gfUserHeader[k+1].toCompile;
        word[11] = gfUserHeader[k+1].attributes;
        for (m = 0; m < 12; m+=2)  {
//first word
            *flashPtr = ~word[m]; //~for better compatibility with empty SRAM
            flashPtr++;
//second word
            *flashPtr = ~word[m+1]; //~for better compatibility with empty SRAM
            flashPtr++;
//burn
            while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
            FLASH->SR |= FLASH_SR_EOP;
            if (FLASH->SR & FLASH_SR_OPERR)  return 0;
        }
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

//erase Page26 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((26 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write first part of gTCode into Page26 + offset
    flashPtr = (uint32_t *)(PAGE26 + 0x800 * offset);
    FLASH->CR |= FLASH_CR_PG;
    for (k = 0; k < 0x200; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k]; //~for better compatibility with empty SRAM
	flashPtr++;
//second word
	*flashPtr= ~gfTCode[k+1];//~for better compatibility with empty SRAM
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page27 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((27 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write second part of gTCode into Page27 + offset
    flashPtr = (uint32_t *)(PAGE27 + 0x800 * offset);
    FLASH->CR |= FLASH_CR_PG;

    for (k = 0x200; k < 0x400; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k]; //~for better compatibility with empty SRAM
	flashPtr++;
//second word
	*flashPtr= ~gfTCode[k+1];//~for better compatibility with empty SRAM
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page28 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((28 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    FLASH->CR &= ~FLASH_CR_PER;

//write third part of gTCode into Page28
    flashPtr = (uint32_t *)(PAGE28 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x400; k < 0x600; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page29 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((29 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write fourth part of gTCode into Page29 + offset
    flashPtr = (uint32_t *)(PAGE29 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x600; k < 0x800; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr= ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page30 + ?ffset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((30 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
     FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR) return 0;
    FLASH->CR &= ~FLASH_CR_PER;

//write fifth part of gTCode into Page30 + offset
    flashPtr = (uint32_t *)(PAGE30 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x800; k < 0xA00; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page31 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((31 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write sixth part of gTCode into Page31 + offset
    flashPtr = (uint32_t *)(PAGE31 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0xA00; k < 0xC00; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    sendTB(SPACE);


//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
    return 1;
}


//###########################################################################

//USART Support
//*************

#ifdef STLINK_CUTOFF
void initUsart1(void) // ****************************************************
// USART1-TxD is PortB,6
// USART1-RxD is PortA,10
// **************************************************************************
{
uint16_t baudset;

    RCC->APB2ENR |=  RCC_APB2ENR_USART1EN ;   // USART1 clock enable
    msDelay(5);

//STM32L476 port default mode = ANALOG !!
	GPIOA->MODER &= ~GPIO_MODER_MODE10_0;   //PB6 and PA10 alternate function
	GPIOB->MODER &= ~GPIO_MODER_MODE6_0;    //10=Alternate function mode
	GPIOA->MODER |= GPIO_MODER_MODE10_1;
	GPIOB->MODER |= GPIO_MODER_MODE6_1;

//USART1 is AF7
    GPIOA->AFR[1] &= 0xFFFFF0FF;        //PA10 is RX
    GPIOA->AFR[1] |= 0x00000700;
    GPIOB->AFR[0] &= 0xF0FFFFFF;        //PB6 is TX
    GPIOB->AFR[0] |= 0x07000000;

    gUSART_RXHEAD = gUSART_RXBUF;
    gUSART_RXTAIL = gUSART_RXBUF;

    msDelay(5);

    switch(gBaud) {
    case 96:
        baudset = 0x1380;       //9600Baud: 312.0 at 48 MHz APB2 Clock
        break;
    case 19:
        baudset = 0x09C0;       //19200Baud: 156.0 at 48 MHz APB2 Clock
        break;
    case 31:
        baudset = 0x0600 ;      //MIDI=31250Baud: 96.0 at 48 MHz APB2 Clock
        break;
    case 38:
        baudset = 0x04E0 ;      //38400Baud: 78.0 at 48 MHz APB2 Clock
        break;
    case 57:
        baudset = 0x0340;       //57600Baud: 52.0 at 48 MHz APB2 Clock
        break;
    default:
    case 11:
        baudset = 0x01A0;       //default (virgin flash)
                                //115200Baud: 26.0 at 48 MHz APB2 Clock
    }
    USART1->BRR = baudset; //default 115200Baud: 17.36 at 24 MHz APB2 Clock
#ifdef EXTERN_TXINVERT
	USART1->CR2 = 0;		//1 Stopbit-actual hardware with external inverter
#else
	USART1->CR2 = USART_CR2_TXINV;	//1 Stopbit, TX level inverted by CPU
#endif
    USART1->CR3 = 0;			//default:no Handshake
    USART1->CR1 = USART_CR1_UE | USART_CR1_RXNEIE |USART_CR1_TE |USART_CR1_RE;
													//8Bit, no Parity, RX Int
	gBaudSet = 1;
    NVIC_EnableIRQ(USART1_IRQn);
       //effectively TX interrupt is enabled when bytes are put into the queue
}


void sendB(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
uint32_t fillstate;

    while(!(USART1->ISR&USART_ISR_TXE));
    USART1->TDR = data;
}


void sendTB(uint8_t data) // ************************************************
//send one byte via serial terminal port (USB and/or USART)
//0=USB and USART both terminal, 1+3= USB is terminal, 2+4=USART is terminal
// **************************************************************************
{
    if (gSplit == 0)  {
        sendB(data);
#ifndef USB_DEBUG
        if (guConfigured)  sendUsbByte(data);
#endif
    }
#ifndef USB_DEBUG
    else if (((gSplit == 1) || (gSplit == 3)) && (guConfigured))  {
                            //USB is terminal(virtual COM),USART is serial I/O
        sendUsbByte(data);
    }
#endif
    else  sendB(data);      //USART is terminal, USB is serial or MIDI I/O
}


void sendIOB(uint8_t data) // ************************************************
//send one byte via serial I/O
// **************************************************************************
{
    if ((gSplit == 1) || (gSplit == 3)) { 	    //USB is terminal(virtual COM),
        sendB(data); 						    //USART is serial I/O
    }
#ifndef USB_DEBUG
    else if ((gSplit == 2) && (guConfigured)) { //USART=terminal, USB=serial I/O
        sendUsbByte(data);
    }
    else if ((gSplit == 4) && (guConfigured))  { //USART=terminal, USB=MIDI I/O
        enCodeMidi(data);
    }
#endif
	else  gfErrors |= ERR_CONTEXT;		//gSplit=0
}

	
void sendNibble(uint8_t data) // ********************************************
//0=USB and USART both terminal, 1= USB is terminal, 2||3=USART is terminal
//this communication is exclusively performed with "terminal"
// **************************************************************************
{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
    if (gSplit == 0)  {
        sendB(asciiNibble);
#ifndef USB_DEBUG
        if (guConfigured)  sendUsbByte(asciiNibble);
#endif
    }
#ifndef USB_DEBUG
    else if (((gSplit == 1) || (gSplit == 3)) && (guConfigured))  {
                            //USB is terminal(virtual COM),USART is serial I/O
        sendUsbByte(asciiNibble);
    }
#endif
    else sendB(asciiNibble);
}

#else

void initUsart2(void) // ****************************************************
// USART2-TxD is PortA,2
// USART2-RxD is PortA,3
// **************************************************************************
{
    RCC->APB1ENR1 |=  RCC_APB1ENR1_USART2EN ;   // USART2 clock enable

	GPIOA->MODER &= ~(GPIO_MODER_MODE2_0 | GPIO_MODER_MODE3_0);	
											//pin 2,3  alternate function mode
	GPIOA->MODER |= (GPIO_MODER_MODE2_1 | GPIO_MODER_MODE3_1); 
											//10=Alternate function mode
//USART2 is AF7
    GPIOA->AFR[0] &= 0xFFFF00FF;
    GPIOA->AFR[0] |= 0x00007700;

    gUSART_RXHEAD = gUSART_RXBUF;
    gUSART_RXTAIL = gUSART_RXBUF;

    USART2->BRR = 0x01A0; //default 115200Baud: 26.0 at 48 MHz APB2 Clock
                          //only option when used together with onboard ST-Link

    USART2->CR2 = 0;			//default:1 Stopbit
    USART2->CR3 = 0;			//default:kein Handshake
    USART2->CR1 = USART_CR1_UE | USART_CR1_RXNEIE |USART_CR1_TE | USART_CR1_RE;
                                //8Bit, no Parity, RX Int
    NVIC_EnableIRQ(USART2_IRQn);
        //effectively TX interrupt is enabled when bytes are put into the queue
}


void sendTB(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
uint32_t fillstate;

    while(!(USART2->ISR&USART_ISR_TXE));
    USART2->TDR = data;
}


void sendNibble(uint8_t data) // ********************************************
// **************************************************************************
{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
    sendTB(asciiNibble);
}
#endif

void sendString(const uint8_t * str) // *************************************
// **************************************************************************
{
 uint8_t byte;

    while  ((byte = *str++) != 0)  sendTB(byte);
}


void sendX8(uint32_t data) // *********************************************
// send bits 0...7 in hex format (without leading 0x)
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendTB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendTB(temp);
    sendTB(0x20);
}


void sendX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
    sendTB(0x20);

}

void sendNX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble, noLead;
    noLead = 0;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    if (nibble)  {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
     if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
}


void send0X32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
//	if (data >= 10)  {			//modified 11Jan24
		sendTB('0');
		sendTB('x');
//	}
	sendNX32(data);
}


void sendS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendNS32(data);
    sendTB(0x20);
}


 void sendLS32(uint32_t data) // ********************************************
//send signed decimal number string. Leading zeroes are suppressed
//leading SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendTB(0x20);
    sendNS32(data);
}


void sendNS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//no leading and trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
uint32_t quotient;

    if (data >= 0x80000000) {
        sendTB('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    if (data < 100000) goto TH10;
    if (data < 1000000) goto TH100;
    if (data < 10000000) goto MILLIONS;
    if (data < 100000000) goto MIL10;
    if (data < 1000000000) goto MIL100;
    quotient = data/1000000000;
    data = data - quotient*1000000000;
    sendTB((uint8_t)quotient+ 0x30);
MIL100:
    quotient = data/100000000;
    data = data - quotient*100000000;
    sendTB((uint8_t)quotient + 0x30);
MIL10:
    quotient = data/10000000;
    data = data - quotient*10000000;
    sendTB((uint8_t)quotient + 0x30);
MILLIONS:
    quotient = data/1000000;
    data = data - quotient*1000000;
    sendTB((uint8_t)quotient + 0x30);
TH100:
    quotient = data/100000;
    data = data - quotient*100000;
    sendTB((uint8_t)quotient+ 0x30);
TH10:
    quotient = data/10000;
    data = data - quotient*10000;
    sendTB((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendTB((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendTB((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendTB((uint8_t)quotient + 0x30);
ONES:
    sendTB((uint8_t)data + 0x30) ;
}

#ifdef STLINK_CUTOFF
uint32_t pullUsartRxBuf(void) // ********************************************
#else
uint32_t pullTermRxBuf(void) // *********************************************
#endif
// pull 1 byte asynchronously from USART reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint32_t recv;

	servSystem();   //serve watchdog etc.

    if (gUSART_RXTAIL!=gUSART_RXHEAD)  {
        recv= (uint32_t)((*gUSART_RXTAIL++ & 0x000000FF)|0x00008000);
        if (gUSART_RXTAIL==gUSART_RXBUFTOP) {
                gUSART_RXTAIL=gUSART_RXBUF;
        }
#ifdef STLINK_CUTOFF
        GPIOC->BSRR = GPIO_BSRR_BR_0; // clear bit PC0,LED OFF
#else
        GPIOA->BSRR = GPIO_BSRR_BR_5; // clear bit PA5,LED OFF
#endif
        gLEDCOUNT = 4; //ca 300 milliseconds
        return recv;
    }
    else return 0;
}

#ifdef STLINK_CUTOFF
uint32_t pullTermRxBuf(void)  // *******************************************
// receive byte from serial terminal port with priority order USB > RS232
// **************************************************************************
{
uint32_t recv;

    if (gSplit == 0)  {
        if (guConfigured)  {            //ASART and USB are parallel terminal
            recv = pullUsbRxBuf();
            if (recv) return recv;      //USB has priority over USART
        }
        recv = pullUsartRxBuf();
        return recv;
    }
    else if (((gSplit == 1)||(gSplit == 3))&&(guConfigured)) {//USB is terminal
        recv = pullUsbRxBuf();  //if (guConfigured == 0): no terminal available
        return recv;
    }
    else   {                            //gSplit = 2 or 4:USART is terminal
            recv = pullUsartRxBuf();
            return recv;
    }
    return 0;   //formally, impossible in practice
}


uint32_t pullIORxBuf(void)  // *********************************************
// receive byte from serial terminal port with priority order USB > RS232
// *************************************************************************
{
uint32_t recv;

    switch (gSplit)  {
    case 1:
        recv = pullUsartRxBuf();  //USART is serial I/O (COM)
        return recv;
    case 3:
        recv = pullUsartRxBuf();  //USART is MIDI IN
        if (recv)  MidiScanner(recv & 0x000000FF);
        return 0;               //no further action in "query[]";
    case 2:                     //USB is serial I/O as virtual COM
        if (guConfigured)  {
            recv = pullUsbRxBuf();
            return recv;
        }
        else return 0;          //if(guConfigured==0) no serial I/O available
    case 4:                     //USB is serial I/O as MIDI interface
        if (guConfigured)  {
            recv = pullUsbRxBuf();
            if (recv)  MidiScanner(recv & 0x000000FF);
            return 0;           //no further action in "query[]";
        }
        else return 0;          //if(guConfigured==0) no serial I/O available
    case 0:                     //if gSplit==0: no serial I/O available
    default:
        return  0;
    }
}


void  MidiScanner(uint32_t byte)  // ****************************************
// accepts only MIDI channel as stored in "gChannel", 
// exception: gChannel = 0xFF, then all MIDI channels are accepted
// **************************************************************************
{
static uint32_t status, data1, msgtype;
static uint32_t progress;

    if (byte >= 0x80) {					//new status byte
			if ((byte == 0xF0) || (byte == EOX) || (byte == 0xF4) ||
						(byte == 0xF5) || (byte == 0xF9) ||(byte == 0xFD) ) {
//								kill SysEx(F0),EOX is end of SysEx message,
//								0xF4, 0xF5, 0xF9 and 0xFD are not defined
				progress = 0;			//reset parser
				status = 0;				//drop input until next status byte
				msgtype = 0;
				return;
			}
			if ((byte == 0xF6) || (byte >= 0xF8))  { //1 byte msg
				gMIDIBUF[gMIDI_HEAD] = byte << 16;		//data bytes = 0
				gMIDI_HEAD++;
				if (gMIDI_HEAD == MIDI_BUFLEN) gMIDI_HEAD = 0;
				progress = 0;			//reset parser
				status = 0;				//drop input until next status byte
				msgtype = 0;
				return;
			}
// "normal" MIDI Channel Msg:
			if (((byte & 0x0000000F) != (gChannel - 1)) && (gChannel != 0)) {
				progress = 0;			//reset parser
				status = 0;				//drop input until next status byte
				msgtype = 0;
				return;
			}	//else accepted status byte:
			status = byte;
			msgtype = byte & 0xF0;
			progress = 1;
	}
	else  {								//new data byte
		if (progress == 0)  {	//SysEx or byte out of context, reset parser
			status = 0;
			msgtype = 0;
			data1 = 0;
			return;
		}
		if (progress == 1)  {
			data1 = byte;
			if ((msgtype == 0xC0)||(msgtype == 0xD0) || (status == 0xF1)||
									(status == 0xF3)) { //2 byte msgs
				gMIDIBUF[gMIDI_HEAD] = (status << 16) | data1;
				gMIDI_HEAD++;					//bits 0..7 for easier parsing
				if (gMIDI_HEAD == MIDI_BUFLEN) gMIDI_HEAD = 0;
				return;		//keep progress==1 and gStatus for Running State
			}
			progress = 2;
			return;
		}
		if (progress == 2)  {					//any 3 byte messages
			gMIDIBUF[gMIDI_HEAD] = (status << 16) | (data1 << 8) | byte;
			gMIDI_HEAD++;					//bits 0..7 for easier parsing
			if (gMIDI_HEAD == MIDI_BUFLEN) gMIDI_HEAD = 0;
			progress = 1;   //reset progress and keep status for Running State
		}
	}
}


uint32_t pullMidiBuf(void) // ********************************************
// pull one word asynchronously from MIDIBUF
// check first if something received. If yes then set bit 31. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint32_t recv;

	servSystem();   //serve watchdog etc.

	if (gMIDI_TAIL != gMIDI_HEAD)  {
		recv= gMIDIBUF[gMIDI_TAIL] | 0x80000000;
		gMIDI_TAIL++;
		if (gMIDI_TAIL== MIDI_BUFLEN)  gMIDI_TAIL= 0;
		return recv;
	}
	else return 0;
}
#endif

// ##########################################################################
// Interrupt Handlers

void SysTick_Handler(void)  // **********************************************
// **************************************************************************
{
	if (gSYSCOUNT)  gSYSCOUNT--;		//controls servSystem()
    if (gDELAYCOUNT) gDELAYCOUNT--;
    if (gTIMEOUT) gTIMEOUT--;			//to be used in any subroutine
    if (gfMSCOUNT) gfMSCOUNT --;		//used by op_MS,OP_DS
    if (gfTIXCOUNT) gfTIXCOUNT --;		//used by op_TIX,op_WAIT
	if (gfSTEPCOUNT1) gfSTEPCOUNT1--;	//stepper timing
	if (gfSTEPCOUNT2) gfSTEPCOUNT2--;
	if (gfSTEPCOUNT3) gfSTEPCOUNT3--;
	gfRandom ^= gfRandom << 13;				//xorshift32 algorithm/white noise
	gfRandom ^= gfRandom >> 17;
	gfRandom ^= gfRandom << 5;	
}

#ifdef STLINK_CUTOFF
void USART1_IRQHandler(void)    // *****************************************
// *************************************************************************
{
uint32_t recv;

	if (USART1->ISR & USART_ISR_RXNE)  {
		recv = USART1->RDR & 0x00FF;
		if ((gSplit != 1) && (gSplit != 3)) {
							//	USART is terminal and MiniDmx is not active
			if (recv == CTRL_B) gfTermCtrlBD = CTRL_B;
			if (recv != CTRL_D)  {
				*gUSART_RXHEAD = (uint8_t)recv;
				gUSART_RXHEAD++;
				if (gUSART_RXHEAD==gUSART_RXBUFTOP)gUSART_RXHEAD=gUSART_RXBUF;
			}
			else  op_DEBUG();
		}
		else {									//no special char handling
			*gUSART_RXHEAD = (uint8_t)recv;
			gUSART_RXHEAD++;
			if (gUSART_RXHEAD==gUSART_RXBUFTOP)  gUSART_RXHEAD=gUSART_RXBUF;
		}
	}
}

#else

void USART2_IRQHandler(void)    // *****************************************
// *************************************************************************
{
uint32_t test;

	if (USART2->ISR & USART_ISR_RXNE)  {
		test = USART2->RDR & 0x00FF;
		if (test == CTRL_B) gfTermCtrlBD = CTRL_B;
		if (test != CTRL_D)  {
			*gUSART_RXHEAD = (uint8_t)test;
			gUSART_RXHEAD++;
			if (gUSART_RXHEAD==gUSART_RXBUFTOP)  gUSART_RXHEAD=gUSART_RXBUF;
		}
		else  op_DEBUG();
	}
}
#endif
